use crate::{ip::IpData, ipblc::PKG_NAME};

use std::{
    io::Error,
    net::{Ipv4Addr, Ipv6Addr},
};

use rustables::*;

pub enum FwTableType {
    IPv4,
    IPv6,
}

pub enum FwAction {
    Add,
    Delete,
}

macro_rules! initrules {
    ($batch:expr, $table:expr, $chain:ident, $reset:expr) => {
        $chain.set_hook(Hook::new(HookClass::In, 1));

        $batch.add(&$chain, MsgType::Add);
        if $reset {
            $batch.add(&Rule::new(&$chain).unwrap(), MsgType::Del);
        }
    };
}

macro_rules! makerules {
    ($ipdata:ident, $chain:ident, $batch:ident, $t:ty, $ip_t:ident,$action:ty) => {
        let ip = $ipdata.ip.parse::<$t>().unwrap();
        Rule::new(&$chain)
            .unwrap()
            .saddr(ip.into())
            .drop()
            .add_to_batch(&mut $batch);
    };
}

pub fn fwglobalinit(t: FwTableType, reset: bool) -> (Batch, Chain) {
    let table_name: String;
    let table: Table;
    let mut chain: Chain;
    match t {
        FwTableType::IPv4 => {
            table_name = format!("{PKG_NAME}4");
            table = Table::new(ProtocolFamily::Ipv4).with_name(table_name);
            chain = Chain::new(&table)
                .with_policy(ChainPolicy::Accept)
                .with_name(PKG_NAME);
        }
        FwTableType::IPv6 => {
            table_name = format!("{PKG_NAME}6");
            table = Table::new(ProtocolFamily::Ipv6).with_name(table_name);
            chain = Chain::new(&table)
                .with_policy(ChainPolicy::Accept)
                .with_name(PKG_NAME);
        }
    }
    let mut batch = Batch::new();

    batch.add(&table, MsgType::Add);
    initrules!(batch, table, chain, reset);

    (batch, chain)
}

pub fn fwblock<'a>(ip_add: &IpData) -> std::result::Result<(), error::QueryError> {
    let (mut batch4, chain4) = fwglobalinit(FwTableType::IPv4, false);
    let (mut batch6, chain6) = fwglobalinit(FwTableType::IPv6, false);

    match ip_add.t {
        4 => {
            makerules!(ip_add, chain4, batch4, Ipv4Addr, ipv4, FwAction::Add);
        }
        6 => {
            makerules!(ip_add, chain6, batch6, Ipv6Addr, ipv6, FwAction::Add);
        }
        _ => {}
    }

    // validate and send batch
    for b in [batch4, batch6] {
        match b.send() {
            Ok(_) => {}
            Err(e) => {
                println!("error sending batch: {e}");
                return Err(e);
            }
        };
    }

    Ok(())
}

pub fn fwunblock<'a>(ips_del: &Vec<IpData>) -> std::result::Result<(), Error> {
    let (mut batch4, chain4) = fwglobalinit(FwTableType::IPv4, false);
    let (mut batch6, chain6) = fwglobalinit(FwTableType::IPv6, false);

    // to implement
    /*for ip_del in ips_del {
        match ip_del.t {
            4 => {
                makerules!(ip_del, chain4, batch4, Ipv4Addr, ipv4, FwAction::Del);
            }
            6 => {
                makerules!(ip_del, chain6, batch6, Ipv6Addr, ipv6, FwAction::Del);
            }
            _ => {}
        }
    }*/
    Ok(())
}

pub fn get_current_rules(table_name: &str, chain_name: &str) -> Result<(), Error> {
    let get_table = || -> Result<Option<Table>, Error> {
        let tables = list_tables().unwrap();
        for table in tables {
            if let Some(name) = table.get_name() {
                println!("Found table {}", name);

                if *name == table_name {
                    return Ok(Some(table));
                }
            }
        }

        Ok(None)
    };

    let get_chain = |table: &Table| -> Result<Option<Chain>, Error> {
        let chains = list_chains_for_table(table).unwrap();
        for chain in chains {
            if let Some(name) = chain.get_name() {
                println!("Found chain {}", name);

                if *name == chain_name {
                    return Ok(Some(chain));
                }
            }
        }

        Ok(None)
    };

    let table = get_table()?.expect("no table?");
    let chain = get_chain(&table)?.expect("no chain?");

    let rules = list_rules_for_chain(&chain).unwrap();
    for mut rule in rules {
        println!("{:?}", rule);
    }
    Ok(())
}

pub fn fw_rules_count() -> i64 {
    0
}
